import token_html from "./get_copilot_token.html";

export default {
  async fetch (request, env) {
    if (request.method === "OPTIONS") {
      return handleOPTIONS(request);
    }
    const url = new URL(request.url);
    if (url.pathname === "/token") {
      return new Response(token_html, {
        headers: { "Content-Type": "text/html" }
      });
    }
    if (url.pathname === "/corsproxy" && url.search.startsWith("?https://")) {
      let resp = await fetch(url.search.substring(1), request);
      resp = new Response(resp.body, resp);
      resp.headers.set("Access-Control-Allow-Origin", "*");
      return resp;
    }
    const auth = request.headers.get("Authorization");
    let authKey = auth && auth.split(" ")[1];
    if (!authKey) {
      return new Response("401 Unauthorized", {
        status: 401
      });
    }
    const tokenCache = env.KV;
    let token, endpoint;
    try {
      token = await tokenCache.get(authKey);
      if (token) {
        endpoint = await tokenCache.get(authKey + "-ep");
      }
    } catch (e) {
      console.error(e);
    }
    if (!token) {
      let expires_at, endpoints, errResponse;
      ({ token, expires_at, endpoints, errResponse } = await getToken(authKey));
      if (errResponse) { return errResponse; }
      endpoint = endpoints?.api;
      if (expires_at) {
        try {
          await tokenCache.put(authKey, token, { expiration: expires_at });
          if (endpoint) {
            await tokenCache.put(authKey + "-ep", endpoint, { expiration: expires_at });
          }
        } catch (e) {
          console.error(e);
        }
      }
    }
    return makeRequest(request, url.pathname, token, endpoint || "https://api.githubcopilot.com");
  }
};

const handleOPTIONS = async () => {
  return new Response(null, {
    headers: {
      "Access-Control-Allow-Origin": "*",
      "Access-Control-Allow-Methods": "*",
      "Access-Control-Allow-Headers": "*",
    }
  });
};

const chatVersion = "0.29.0";
const vscodeVersion = "1.102.0";
const apiVersion = "2025-04-01";
const githubapihost = "api.github.com";
const getToken = async (authKey) => {
  if (authKey.startsWith("sk-")) { // dumb attempt to hide real auth key from malicious web services
    const decoded = atob(authKey.substring(3));
    if (/^[ -~]+$/.test(decoded)) { // ascii
      authKey = decoded;
    }
  }
  let response = await fetch(`https://${githubapihost}/copilot_internal/v2/token`, {
    method: "GET",
    headers: {
      "Authorization": `token ${authKey}`,
      "Host": githubapihost,
      "Editor-Version": `vscode/${vscodeVersion}`,
      "Editor-Plugin-Version": `copilot-chat/${chatVersion}`,
      "User-Agent": `GitHubCopilotChat/${chatVersion}`,
      "Accept": "*/*",
      "Accept-Encoding": "*",
      //'Accept-Encoding': 'gzip, deflate, br',
    },
  });
  if (!response.ok) {
    return { errResponse: response };
  }
  const text = await response.text();
  let data;
  try {
    data = JSON.parse(text);
  } catch (e) {
    console.error(e.message,"\n",text);
    return { errResponse: new Response(e.message + "\n" + text, { status: 400 }) };
  }
  if (!data.token) {
    console.error("token not found:\n", data);
    return { errResponse: new Response("token not found:\n" + text, { status: 400 }) };
  }
  return data;
};

const makeHeaders = async (token, endpoint) => {
  const createSha256Hash = async (input) => {
    const encoder = new TextEncoder();
    const data = encoder.encode(input);
    const hash = await crypto.subtle.digest("SHA-256", data);
    return Array.from(new Uint8Array(hash)).map(b => b.toString(16).padStart(2, "0")).join("");
  };
  return {
    "Authorization": `Bearer ${token}`,
    "Host": endpoint.replace(/^https?:\/\//, ""),
    "X-Request-Id": crypto.randomUUID(),
    "X-Github-Api-Version": apiVersion,
    "Vscode-Sessionid": crypto.randomUUID() + Date.now().toString(),
    "vscode-machineid": await createSha256Hash(token),
    "Editor-Version": `vscode/${vscodeVersion}`,
    "Editor-Plugin-Version": `copilot-chat/${chatVersion}`,
    "Openai-Organization": "github-copilot",
    "Copilot-Integration-Id": "vscode-chat",
    "Openai-Intent": "conversation-panel",
    "Content-Type": "application/json",
    "User-Agent": `GitHubCopilotChat/${chatVersion}`,
    "Accept": "*/*",
    "Accept-Encoding": "gzip, deflate, br",
  };
};

const makeRequest = async (request, path, token, endpoint) => {
  if (path.startsWith("/v1/")) {
    path = path.substring(3);
  }
  const response = await fetch(`${endpoint}${path}`, {
    method: request.method,
    headers: await makeHeaders(token, endpoint),
    body: request.body,
  });
  const headers = new Headers(response.headers);
  headers.set("Access-Control-Allow-Origin", "*");
  let body;
  if (response.ok && path === "/chat/completions" && request.method === "POST") {
    if (headers.get("Transfer-Encoding") === "chunked") { // is stream
      body = response.body
        .pipeThrough(new TextDecoderStream())
        .pipeThrough(new TransformStream({ transform: cleanStream, buffer: "" }))
        .pipeThrough(new TextEncoderStream());
      headers.set("Content-Type", "text/event-stream");
    } else {
      body = clean(await response.text());
    }
  } else {
    body = response.body;
  }
  return new Response(body, { status: response.status, statusText: response.statusText, headers });
};

const clean = (str) => {
  let json;
  try {
    json = JSON.parse(str);
  } catch (e) {
    console.error(e);
    return str;
  }
  json.object = "chat.completion";
  delete json.prompt_filter_results;
  for (const item of json.choices) {
    delete item.content_filter_results;
    //item.logprobs = null;
  }
  return JSON.stringify(json);
};

const cleanLine = (str) => {
  if (str.startsWith("data: ")) {
    let json;
    try {
      json = JSON.parse(str.substring(6));
    } catch (e) {
      if (str !== "data: [DONE]") {
        console.error(e);
      }
      return str;
    }
    if (json.choices.length === 0) { return; } // json.prompt_filter_results
    json.object = "chat.completion.chunk";
    for (const item of json.choices) {
      delete item.content_filter_offsets;
      delete item.content_filter_results;
      const delta = item.delta;
      for (const k in delta) {
        if (delta[k] === null ) { delete delta[k]; }
      }
      //delta.content = delta.content || "";
      //item.logprobs = null;
      //item.finish_reason = item.finish_reason || null;
    }
    return "data: " + JSON.stringify(json);
  }
};

const delimiter = "\n\n";
async function cleanStream (chunk, controller) {
  chunk = await chunk;
  if (!chunk) {
    if (this.buffer) {
      controller.enqueue(cleanLine(this.buffer) || this.buffer);
    }
    controller.enqueue("\n");
    controller.terminate();
    return;
  }
  this.buffer += chunk;
  let lines = this.buffer.split(delimiter);
  for (let i = 0; i < lines.length - 1; i++) {
    const line = cleanLine(lines[i]);
    if (line) {
      controller.enqueue(line + delimiter);
    }
  }
  this.buffer = lines[lines.length - 1];
}
